"""Ready-to-use hyperparameter search module.""" 

import csv
from collections import OrderedDict
import itertools
import numpy as np
import os
import pandas as pd
from pathlib import Path
import random
import shortuuid
import sys
from tqdm import tqdm

from dlkoopman.state_pred import StatePred, StatePredDataHandler
from dlkoopman.traj_pred import TrajPred, TrajPredDataHandler


def _gen_colnames(perf, do_val=True) -> list[str]:
    """
    Inputs:
        perf (str) - Either of <recon/lin/pred>_<loss/anae> or total_loss
    
    Return a list of strings that are column names of output CSV - see code for details
    """
    ret = [
        f'avg_{perf}_tr',
        f'final_{perf}_tr'
    ]
    if do_val:
        ret.extend([
            f'avg_{perf}_va',
            f'best_{perf}_va',
            f'bestep_{perf}_va'
        ])
    return ret

def _compute_stats(stats, perf, do_val=True, avg_ignore_initial_epochs=0) -> list[float]:
    """
    Inputs:
        stats - A MODEL_CLASS.stats object
        perf (str) - Either of <recon/lin/pred>_<loss/anae> or total_loss
        do_val (bool) - Whether to return validation stats or not
        avg_ignore_initial_epochs (bool) - See doc of run_hyp_search()
    
    Return:
        avg(perf_tr)
        last(perf_tr)
        avg(perf_va) (if do_val is True)
        best(perf_va) (if do_val is True)
        bestep(perf_va) (if do_val is True)
    """
    ret = [
        np.mean(stats[f'{perf}_tr'][avg_ignore_initial_epochs:]),
        stats[f'{perf}_tr'][-1]
    ]
    if do_val:
        ret.extend([
            np.mean(stats[f'{perf}_va'][avg_ignore_initial_epochs:]),
            np.min(stats[f'{perf}_va']),
            np.argmin(stats[f'{perf}_va'])+1
    ])
    return ret



def run_hyp_search(
    dh, hyp_options,
    numruns=None, avg_ignore_initial_epochs=0, sort_key='avg_pred_anae_va'
) -> Path:
    """Perform many runs of a type of predictor model (either `StatePred` or `TrajPred`) with different configurations on given data, and save the metrics for each run.

    Use the results to pick a good predictor configuration.

    **The method can be interrupted at any time and the intermediate results will be saved**.

    ## Saved statistics
    For each of the following `<x>`:

    - 'recon_loss', 'recon_anae' - Reconstruction loss and ANAE.
    - 'lin_loss', 'lin_anae' - Linearity loss and ANAE.
    - 'pred_loss', 'pred_anae' - Predicton loss and ANAE.
    - 'total_loss' - Total loss.

    save the following statistics:

    If only training data is provided:

    - `avg_<x>_tr` - Average training `<x>` across all epochs.
    - `final_<x>_tr` - Final training `<x>` after last epoch.

    If both training and validation data is provided (recommended), save the above, and additionally:

    - `avg_<x>_va` - Average validation `<x>` across all epochs.
    - `best_<x>_va` - Best validation `<x>` across all epochs.
    - `bestep_<x>_va` - Epoch number at which best validation `<x>` was achieved.

    For more details on losses and ANAEs, refer to [metrics](https://galoisinc.github.io/dlkoopman/metrics.html).

    ## Parameters
    - **dh** (*StatePredDataHandler, or TrajPredDataHandler*) - Data handler providing data. **Model to be run is inferred from data, i.e. either `StatePred` or `TrajPred`.**
    
    - **hyp_options** (*dict[str,list]*) - Input arguments to model and its methods will be swept over these values across runs. As an example, when `dh` is a `StatePredDataHandler`:
    ```python
    hyp_options = {
        ## arguments to __init__()
        'rank': [6,8], # must be provided since it's a required argument
        'encoded_size': 50, # must be provided since it's a required argument; this input stays constant across runs
        'encoder_hidden_layers': [[100,100],[200,100,50]]
        
        ## arguments to train_net()
        'numepochs': list(range(100,501,100)), # ranges must be provided as lists
        'clip_grad_norm': 5 # this input stays constant across runs
    }
    
    # this results in 2*2*5=20 possible runs
    ```

    - **numruns** (*int, optional*) - The total number of possible model runs is \\(N =\\) the number of elements in the Cartesian product of the values of `hyp_options` (in the above example, this is 20). If `numruns` is `None` or \\(>N\\), run \\(N\\) runs, otherwise run `numruns` runs.<br>Since each run takes time, it is recommended to set `numruns` to a reasonably smaller value when \\(N\\) is large.

    - **avg_ignore_initial_epochs** (*int, optional*) - When collecting average statistics across epochs, ignore this many initial epochs. This is useful because the loss / ANAE values can be quite high early on, which can skew the averages and make it meaningless to compare runs based on the averages.

    - **sort_key** (*str, optional*) - Results in the final CSV will be sorted in ascending order of this column. For possible options, see **Saved statistics**. Note that sorting only happens at the end, and thus will not take place if execution is interrupted. Set to `None` to skip sorting.

    ## Returns
    **output_folder** (*Path*) - The path to a newly created folder containing:
    
    - Results CSV file = `hyp_search_results.csv`
    - Log file = `hyp_search_log.log`
    - If any of the individual model runs resulted in errors, their log files will be stored as well so that you can take a closer look at what went wrong.

    The results CSV contains:
    ```
    -------------------------------------------------------------------------
    | UUID | <hyperparameters swept over> | <loss results> | <anae results> |
    -------------------------------------------------------------------------
    | run1 |              ...             |       ...      |       ...      |
    | run2 |              ...             |       ...      |       ...      |
    | run3 |              ...             |       ...      |       ...      |
    -------------------------------------------------------------------------
    ```
    """
    ## Define do_val
    do_val = len(dh.Xva) > 0
    
    ## Define perfs
    PERFS = [
        'recon_loss', 'lin_loss', 'pred_loss', 'total_loss',
        'recon_anae', 'lin_anae', 'pred_anae'
    ]

    ## Define possible hyp options and MODEL_CLASS
    if isinstance(dh, StatePredDataHandler):
        REQD_CLASS_KEYS = ['rank', 'encoded_size']
        CLASS_KEYS = ['rank', 'encoded_size', 'encoder_hidden_layers', 'decoder_hidden_layers', 'batch_norm']
        TRAIN_KEYS = ['numepochs', 'early_stopping', 'early_stopping_metric', 'lr', 'weight_decay', 'decoder_loss_weight', 'Kreg', 'cond_threshold', 'clip_grad_norm', 'clip_grad_value']
        MODEL_CLASS = StatePred
    elif isinstance(dh, TrajPredDataHandler):
        REQD_CLASS_KEYS = ['encoded_size']
        CLASS_KEYS = ['encoded_size', 'encoder_hidden_layers', 'decoder_hidden_layers', 'batch_norm']
        TRAIN_KEYS = ['numepochs', 'batch_size', 'early_stopping', 'early_stopping_metric', 'lr', 'weight_decay', 'decoder_loss_weight', 'clip_grad_norm', 'clip_grad_value']
        MODEL_CLASS = TrajPred
    else:
        raise ValueError(f"Invalid 'dh' of type {type(dh)}")
    ALL_KEYS = CLASS_KEYS + TRAIN_KEYS

    ## Check that required hyp options are provided
    msg = ""
    for k in REQD_CLASS_KEYS:
        if k not in hyp_options:
            msg += f"'{k}' is a required argument to {MODEL_CLASS.__name__}, so 'hyp_options' must include a key-value pair for {{'{k}': <{k}_values>}}"
    if msg: 
        raise ValueError(msg)

    ## Order key-value pairs into a new dict where CLASS_KEYS comes first, then TRAIN_KEYS
    hyp_options_ordered = OrderedDict()

    USED_CLASS_KEYS = []
    for k in CLASS_KEYS:
        if k in hyp_options:
            USED_CLASS_KEYS.append(k)
            v = hyp_options[k]
            hyp_options_ordered[k] = [v] if type(v) not in [list,set,tuple] else v

    USED_TRAIN_KEYS = []
    for k in TRAIN_KEYS:
        if k in hyp_options:
            USED_TRAIN_KEYS.append(k)
            v = hyp_options[k]
            hyp_options_ordered[k] = [v] if type(v) not in [list,set,tuple] else v

    ## Report invalid hyps
    invalid_hyps = set(hyp_options.keys()).difference(set(hyp_options_ordered.keys()))
    for k in invalid_hyps:
        print(f"WARNING: Key '{k}' in 'hyp_options' is not a valid argument to {MODEL_CLASS.__name__} or any of its methods. This key will be ignored. Keys must be in {ALL_KEYS}.")

    ## Delete hyp_options as workflow has completely moved over to hyp_options_ordered
    del hyp_options

    ## Ignore early stopping metric if early stopping is only False
    if hyp_options_ordered.get('early_stopping', [0]) == [0] and 'early_stopping_metric' in hyp_options_ordered:
        del hyp_options_ordered['early_stopping_metric']
        USED_TRAIN_KEYS.remove('early_stopping_metric')

    ## Get hyps_list_all
    hyps_list_all = list(itertools.product(*[hyp_options_ordered[key] for key in hyp_options_ordered]))
    if numruns and len(hyps_list_all) > numruns:
        hyps_list_all = random.sample(hyps_list_all, numruns)

    ## Process results folder and its files
    output_folder = Path(f'./hyp_search_{shortuuid.uuid()}').resolve()
    output_folder.mkdir()
    results_file = output_folder.joinpath('hyp_search_results.csv')
    log_file = output_folder.joinpath('hyp_search_log.log')

    ## Print initial info
    print('\n********************************************************************************')
    print(f'Starting {MODEL_CLASS.__name__} hyperparameter search. Results will be stored in {results_file}.\n')
    print(f'Performing total {len(hyps_list_all)} runs. You can interrupt the script at any time (e.g. Ctrl+C), and intermediate results will be available in the above file.\n')
    print(f'Log of the entire hyperparameter search, as well as logs of failed {MODEL_CLASS.__name__} runs will also be stored in the same folder.\n')
    print("Hyperparameters' sweep ranges:")
    for k,v in hyp_options_ordered.items():
        print(f"{k} : {', '.join([str(vv) for vv in v])}")
    print('********************************************************************************\n')

    ## Change directory
    original_cwd = os.getcwd()
    os.chdir(output_folder)

    ## Write outputs to log file from here on
    original_stdout = sys.stdout
    sys.stdout = open(log_file, 'w')

    ## Open CSV and write header row
    with open(results_file, 'w') as csvfile:
        csvwriter = csv.writer(csvfile)
        header_row = [
            'UUID',
            *list(hyp_options_ordered.keys())
        ]
        for perf in PERFS:
            header_row += _gen_colnames(perf=perf, do_val=do_val)
        csvwriter.writerow(header_row)

        ## Do runs
        for hyps_list in tqdm(hyps_list_all):
            class_hyps = OrderedDict(zip(USED_CLASS_KEYS, hyps_list[:len(USED_CLASS_KEYS)]))
            train_hyps = OrderedDict(zip(USED_TRAIN_KEYS, hyps_list[len(USED_CLASS_KEYS):]))
            all_hyps = {**class_hyps, **train_hyps}
            print(f'Hyperparameters: {all_hyps}')

            model = MODEL_CLASS(
                dh = dh,
                **class_hyps
            )

            row = [
                model.uuid,
                *list(class_hyps.values()),
                *list(train_hyps.values()),
            ]

            model.train_net(**train_hyps)

            if model.error_flag:
                row += (len(header_row)-len(row))*[np.nan]
                print('WARNING: Encountered error. The log file of this model will be saved for you to take a closer look.')

            else:
                if avg_ignore_initial_epochs >= len(model.stats['total_loss_tr']): # can pick any key
                    print(f"WARNING: The value of `avg_ignore_initial_epochs` = {avg_ignore_initial_epochs} is greater than the actual number of epochs for which the current model instance has run for = {len(model.stats['total_loss_tr'])}. Thus, 'avg' statistics will be equal to the last epoch's value.")
                    _avg_ignore_initial_epochs = len(model.stats['total_loss_tr'])-1
                else:
                    _avg_ignore_initial_epochs = avg_ignore_initial_epochs
                for perf in PERFS:
                    row += _compute_stats(stats=model.stats, perf=perf, do_val=do_val, avg_ignore_initial_epochs=_avg_ignore_initial_epochs)
                os.system(f"rm -rf {model.log_file}")

            csvwriter.writerow(row)
            print()

    ## Back to writing in terminal
    sys.stdout = original_stdout

    ## Delete constant columns and sort results
    df = pd.read_csv(results_file)
    df.drop(columns = [k for k,v in hyp_options_ordered.items() if len(v)==1], inplace=True)
    
    if sort_key:
        try:
            df.sort_values(sort_key, ascending=True, inplace=True)
        except KeyError:
            print(f"WARNING: `sort_key` = '{sort_key}' not found among the columns of the output CSV '{results_file}'. Results will be unsorted.")
        df.to_csv(results_file, index=False)

    ## Change back to original directory
    os.chdir(original_cwd)

    return output_folder



"""Loss and error functions, used to optimize models and report their performance."""


import torch


def anae(ref, new) -> torch.Tensor:
    """Average Normalized Absolute Error (ANAE).
    
    ANAE first normalizes each absolute deviation by the corresponding absolute ground truth, then averages them. This is a useful thing to report since it tells us how much percentage deviation to expect for a new value. E.g. if prediction ANAE for a problem is around 10%, then one can expect a newly predicted value to have an error of around 10% from the actual.
    
    Example: Let
    ```python
    ref = torch.tensor([[-0.1,0.2,0],[100,200,300]])
    new = torch.tensor([[-0.11,0.15,0.01],[105,210,285]])
    ```

    $$ANAE = \\text{Avg}\\left(\\frac{|-0.1-(-0.11)|}{|-0.1|}, \\frac{|0.2-0.15|}{|0.15|}, \\frac{|100-105|}{|100|}, \\frac{|200-210|}{|200|}, \\frac{|300-285|}{|300|}\\right) = 10\\%$$
    Note that:
    
    - Ground truth value of \\(0\\) is ignored. 
    
    - ANAE heavily penalizes deviations for small values of ground truth.

    ## Parameters
    **ref** (*torch.Tensor*) and **new** (*torch.Tensor*) - Error will be calculated between these tensors.

    ## Returns
    **anae** (*torch scalar*) - In percentage.
    """
    anae = torch.abs(ref-new)/torch.abs(ref)
    return 100.*torch.mean(anae[anae!=torch.inf])

def _naae(ref, new) -> torch.Tensor:
    """Normalized Average Absolute Error (NAAE).
    
    NAAE first averages the absolute error, then normalizes that using the average value of the absolute ground truth. This reduces the chance of a few small ground truth values having a huge impact, but loses details by taking the average over all absolute error values, and likewise over all ground truth values.
    
    Example: Let `ref = torch.tensor([[-0.1,0.2,0],[100,200,300]]), new = torch.tensor([[-0.11,0.15,0.01],[105,210,285]])`.
    $$NAAE = \\frac{\\text{Avg}(|0.01|,|0.05|,|-0.01|,|-5|,|-10|,|15|)}{\\text{Avg}(|-0.1|,|0.2|,|0|,|100|,|200|,|300|)} = 5\\%$$

    ## Parameters
    **ref** (*torch.Tensor*) and **new** (*torch.Tensor*) - Error will be calculated between these tensors.

    ## Returns
    **naae** (*torch scalar*) - In percentage.

    ## Caution
    NAAE is experimental, so it is best to stick with ANAE for now.
    """
    return 100.*torch.mean(torch.abs(ref-new))/torch.mean(torch.abs(ref))


def overall_anae(X, Y, Xr, Ypred, Xpred) -> dict[str, torch.Tensor]:
    """Computes overall ANAE for a model.
    
    ## Parameters
    - **X** (*torch.Tensor, shape=(\\*, input_size)*) - Input states, i.e. input to encoder.

    - **Y** (*torch.Tensor, shape=(\\*, encoded_size)*) - Encoded states, i.e. output from encoder, input to decoder.

    - **Xr** (*torch.Tensor, shape=(\\*, input_size)*) - Reconstructed states, i.e. output of decoder.

    - **Ypred** (*torch.Tensor, shape=(\\*, encoded_size)*) - Predicted encoded states obtained from evolving baseline encoded state.

    - **Xpred** (*torch.Tensor, shape=(\\*, input_size)*) - Predicted input states, which are predicted encoded states passed through decoder.

    ## Returns
    **anaes** (*dict[str, torch.Tensor]*)

    - Key **'recon'**: (*torch scalar*) - Reconstruction ANAE between `X` and `Xr`.
    - Key **'lin'**: (*torch scalar*) - Linearity ANAE between `Y` and `Ypred`.
    - Key **'pred'**: (*torch scalar*) - Prediction ANAE between `X` and `Xpred`.
    """
    return {
        'recon': anae(ref=X, new=Xr),
        'lin': anae(ref=Y, new=Ypred),
        'pred': anae(ref=X, new=Xpred)
    }


def overall_loss(X, Y, Xr, Ypred, Xpred, decoder_loss_weight) -> dict[str, torch.Tensor]:
    """Computes overall loss for a model.
    
    ## Parameters
    - **X** (*torch.Tensor, shape=(\\*, input_size)*) - Input states, i.e. input to encoder.

    - **Y** (*torch.Tensor, shape=(\\*, encoded_size)*) - Encoded states, i.e. output from encoder, input to decoder.

    - **Xr** (*torch.Tensor, shape=(\\*, input_size)*) - Reconstructed states, i.e. output of decoder.

    - **Ypred** (*torch.Tensor, shape=(\\*, encoded_size)*) - Predicted encoded states obtained from evolving baseline encoded state.

    - **Xpred** (*torch.Tensor, shape=(\\*, input_size)*) - Predicted input states, which are predicted encoded states passed through decoder.

    - **decoder_loss_weight** (*float, optional*) - Weight the losses between decoder outputs (`recon` and `pred`) by this number. This is to account for the scaling effect of the decoder.

    ## Returns
    **losses** (*dict[str, torch.Tensor]*)

    - Key **'recon'**: (*torch scalar*) - Reconstruction loss between `X` and `Xr`.
    - Key **'lin'**: (*torch scalar*) - Linearity loss between `Y` and `Ypred`.
    - Key **'pred'**: (*torch scalar*) - Prediction loss between `X` and `Xpred`.
    - Key **'total'**: (*torch scalar*) - Total loss = `lin + decoder_loss_weight*(recon+pred)`
    """
    losses = {
        'recon': torch.nn.MSELoss(reduction='mean')(X, Xr),
        'lin': torch.nn.MSELoss(reduction='mean')(Y, Ypred),
        'pred': torch.nn.MSELoss(reduction='mean')(X, Xpred)
    }
    losses['total'] = losses['lin'] + decoder_loss_weight * (losses['recon'] + losses['pred'])
    return losses



"""Utilities"""


import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import random
import torch
from typing import Any


__pdoc__ = {
    'stable_svd': False,
    'moving_avg': False
}


def _safe_inverse(x, epsilon=1e-12):
    return x/(x**2 + epsilon)

class _SVD(torch.autograd.Function):
    @staticmethod
    def forward(self, A):
        U, S, V = torch.svd(A) #NOTE: torch.svd may be depreciated later, requiring the switch to torch.linalg.svd
        self.save_for_backward(U, S, V)
        return U, S, V

    @staticmethod
    def backward(self, dU, dS, dV):
        U, S, V = self.saved_tensors
        Vt = V.t()
        Ut = U.t()
        M = U.size(0)
        N = V.size(0)
        NS = len(S)

        F = (S - S[:, None])
        F = _safe_inverse(F)
        F.diagonal().fill_(0)

        G = (S + S[:, None])
        G.diagonal().fill_(np.inf)
        G = 1/G 

        UdU = Ut @ dU
        VdV = Vt @ dV

        Su = (F+G)*(UdU-UdU.t())/2
        Sv = (F-G)*(VdV-VdV.t())/2

        dA = U @ (Su + Sv + torch.diag(dS)) @ Vt 
        if (M>NS):
            dA = dA + (torch.eye(M, dtype=dU.dtype, device=dU.device) - U@Ut) @ (dU/S) @ Vt 
        if (N>NS):
            dA = dA + (U/S) @ dV.t() @ (torch.eye(N, dtype=dU.dtype, device=dU.device) - V@Vt)
        return dA

def stable_svd(x) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Stable Singular Value Decomposition (courtesy [this](https://github.com/wangleiphy/tensorgrad/blob/master/tensornets/adlib/svd.py), in response to [this](https://github.com/google/jax/issues/2311#issuecomment-984131512)) alternative to [`torch.linalg.svd`](https://pytorch.org/docs/stable/generated/torch.linalg.svd.html), which may encounter NaNs in gradients.

    The other solution is to zero out the NaN gradients as described [here](https://github.com/tensorflow/tensorflow/issues/17476#issue-302663705), however, we don't use this technique.

    ## Parameters
    **x** (*torch.Tensor*) - Matrix whose SVD will be computed. Assume shape to be (m,n).

    ## Returns
    - **U** (*torch.Tensor, shape=(m, min(m,n))*)
    - **Sigma** (*torch.Tensor, shape=(min(m,n),)*)
    - **V** (*torch.Tensor, shape=(n, min(m,n))*)
    """
    return _SVD.apply(x)


def plot_stats(model, perfs=['pred_anae'], start_epoch=1, fontsize=12):
    """Plot stats from a model.

    ## Parameters
    - **model** (*StatePred* or *TrajPred*) - A model with `stats` populated.

    - **perfs** (*list[str]*) - Which variables from `stats` to plot. For each variable, training data and validation data stats are plotted vs epochs, and the title of the plot is the test data stats value.
    
    - **start_epoch** (*int*) - Start plotting from this epoch. Setting this to higher than 1 may be useful when the first few epochs have weird values that skew the y axis scale.

    - **fontsize** (*int*) - Font size of plot title. Other font sizes are automatically adjusted relative to this.

    ## Effects
    Creates plots for each `perf` and saves their png file(s) to `"./plot_<model.uuid>_<perf>.png"`.
    """
    for perf in perfs:
        is_anae = 'anae' in perf

        tr_data = model.stats[perf+'_tr'][start_epoch-1:]
        if model.stats[perf+'_va']:
            va_data = model.stats[perf+'_va'][start_epoch-1:]
            tr_data = tr_data[:len(va_data)] # va_data should normally have size equal to tr_data, but will have lesser size if some error occurred during training. This slicing ensures that only that portion of tr_data is considered which corresponds to va_data.
        epoch_range = range(start_epoch,start_epoch+len(tr_data))

        plt.figure()
        if model.stats[perf+'_te']:
            plt.suptitle(f"Test performance = {model.stats[perf+'_te'][-1]}" + (' %' if is_anae else ''), fontsize=fontsize)

        plt.plot(epoch_range, model.stats[perf+'_tr'][start_epoch-1:], c='MediumBlue', label='Training')
        if model.stats[perf+'_va']:
            plt.plot(epoch_range, model.stats[perf+'_va'][start_epoch-1:], c='DeepPink', label='Validation')

        if is_anae:
            ylim_anae = plt.gca().get_ylim()
            plt.ylim(max(0,ylim_anae[0]), min(100,ylim_anae[1])) # keep ANAE limits between [0,100]
        plt.xlabel('Epochs', fontsize=fontsize)
        plt.ylabel(perf + (' (%)' if is_anae else ''), fontsize=fontsize)
        plt.xticks(fontsize=fontsize-2)
        plt.yticks(fontsize=fontsize-2)

        plt.grid()
        plt.legend(fontsize=fontsize)

        save_path = Path(f'./plot_{model.uuid}_{perf}.png').resolve()
        print(f'Saving figure {save_path}')
        plt.savefig(save_path, dpi=600, bbox_inches='tight', pad_inches=0.1)


def set_seed(seed):
    """Set a random seed to make results reproducible.

    ## Parameters
    **seed** (*int*) - The seed to be set.

    ## Effects
    Sets the random seed to `seed`.
    """
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def _tensorize(arg, dtype, device) -> Any:
    return torch.as_tensor(arg, dtype=dtype, device=device) if arg is not None else torch.tensor([], dtype=dtype, device=device)

def _scale(arg, scale) -> Any:
    return arg/scale

def _shift(arg, shift) -> Any:
    return arg-shift


def _extract_item(v) -> Any:
    """ Given input, return its `.item()` if it can be extracted, otherwise return input. """
    try:
        ret = v.item()
    except (AttributeError, ValueError):
        ret = v
    return ret


def moving_avg(inp, window_size = 3):
    """Compute moving average of the elements of an ordered iterable.

    ## Parameters
    **inp** (*Ordered Array, e.g. list, tuple*) - Input ordered iterable.
    **window_size** (*int*) - How many elements to consider in each moving average.

    ## Return
    out (*Ordered Array, same type as input*) - Moving averaged elements of `inp`.

    Throws a `ValueError` if `window_size` > `len(inp)`.

    ## Example
    ```python
    inp = [1,2,3,5,7]
    out = moving_avg(inp, 3)
    # out = [2, 10/3, 5]
    ```
    """
    if window_size > len(inp):
        raise ValueError(f"'window_size' must be <= length of 'inp', but {window_size} > {len(inp)}")
    return type(inp)(np.convolve(inp, np.ones(window_size), 'valid') / window_size)



"""Configuration options"""


###############################################################################
# Change config options in this section as desired
###############################################################################

precision = "float"
"""Numerical precision of tensors.

**Options** (in increasing order of precision): `"half"` / `"float"` / `"double"`.

**Default**: `"float"`

## Note
Setting `precision = "double"` may make predictions slightly more accurate, however, may also lead to inefficient GPU runtimes.
"""

use_cuda = True
"""If `True`, tensor computations when take place on CuDA GPUs whenever available.

**Options**: `True` / `False`

**Default**: `True`
"""

normalize_Xdata = True
"""If `True`, all input states (training, validation, test) are divided by the maximum absolute value in the training data.

**Options**: `True` / `False`

**Default**: `True`

## Notes

Normalizing data is a generally good technique for deep learning, and is normally done for each feature \\(f\\) in the input data as
$$X_f = \\frac{X_f-\\text{offset}_f}{\\text{scale}_f}$$
(where offset and scale are mean and standard deviation for Gaussian normalization, or minimum value and range for Minmax normalization.) However, *this messes up the spectral techniques such as singular value and eigen value decomposition required in Koopman theory*. Hence, setting `normalize_Xdata=True` will just use a single scale value for normalizing the whole data to get
$$X = \\frac{X}{\\text{scale}}$$
This results in the singular and eigen vectors remaining the same.

## Caution
Setting this to `normalize_Xdata=False` may end the run by leading to imaginary parts of tensors reaching values where the loss function depends on the phase (in `torch >= 1.11`, this leads to `"RuntimeError: linalg_eig_backward: The eigenvectors in the complex case are specified up to multiplication by e^{i phi}. The specified loss function depends on this quantity, so it is ill-defined"`). The only benefit to setting this to `False` is that if the run successfully completes, the final error metrics such as ANAE are slightly more accurate since they are reported on the un-normalized data values.
"""

use_exact_eigenvectors = True
"""If `True`, the exact eigenvectors of the Koopman matrix are used in `StatePred`, if `False`, the projected eigenvectors are used.

**Options**: `True` / `False`

**Default**: `True`

## Notes

For a discussion on exact and projected eigenvectors, see [Tu et al](https://arxiv.org/abs/1312.0041) or [Chapter 1 of Kutz et al](https://epubs.siam.org/doi/pdf/10.1137/1.9781611974508.ch1). The basic idea is that using exact eigenvectors is more accurate, but their computation may become less numerically stable than that of projected eigenvectors.
"""

sigma_threshold = 1e-25
"""When computing the SVD in `StatePred`, singular values lower than this will be reported, since they can be a possible cause of unstable gradients.

**Options**: Any numerical value.

**Default**: `1e-25`
"""

###############################################################################
# End of config options
###############################################################################




###############################################################################
# Do not change anything from here on
###############################################################################


__pdoc__ = {
    'ConfigValidationError': False
}

# Config validation - runs automatically whenever config is imported

class ConfigValidationError(Exception):
    """Raised when config does not validate."""

if precision not in ["half", "float", "double"]:
    raise ConfigValidationError(f'`precision` must be either of "half" / "float" / "double", instead found {precision}')
if use_cuda not in [True, False]:
    raise ConfigValidationError(f'`use_cuda` must be either True or False, instead found {use_cuda}')
if normalize_Xdata not in [True, False]:
    raise ConfigValidationError(f'`normalize_Xdata` must be either True or False, instead found {normalize_Xdata}')
if use_exact_eigenvectors not in [True, False]:
    raise ConfigValidationError(f'`use_exact_eigenvectors` must be either True or False, instead found {use_exact_eigenvectors}')
if type(sigma_threshold) not in [int, float]:
    raise ConfigValidationError(f'`sigma_threshold` must be a number, instead found {sigma_threshold}')


# Set other constants from config values

import torch

_RTYPE = torch.half if precision=="half" else torch.float if precision=="float" else torch.double
_CTYPE = torch.chalf if precision=="half" else torch.cfloat if precision=="float" else torch.cdouble
_DEVICE = torch.device("cuda" if use_cuda and torch.cuda.is_available() else "cpu")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dlkoopman']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.23.0,<2.0.0',
 'pandas>=1.2.3,<2.0.0',
 'shortuuid>=1.0.9,<2.0.0',
 'torch>=1.12.1,<2.0.0',
 'tqdm>=4.61.1,<5.0.0']

setup_kwargs = {
    'name': 'dlkoopman',
    'version': '1.0.2',
    'description': 'A Python package for Koopman theory using deep learning.',
    'long_description': 'A Python package for Koopman theory using deep learning.\n\n\n## Overview\nKoopman theory is a mathematical technique to achieve data-driven approximations of nonlinear dynamical systems by encoding them into a linear space. `dlkoopman` uses deep learning to learn such an encoding, while simultaneously learning the linear dynamics.\n\n### Key features\n- State prediction (`StatePred`) - Train on individual states (snapshots) of a system, then predict unknown states.\n    - E.g: What is the pressure vector on this aircraft for $23.5^{\\circ}$ angle of attack?\n- Trajectory prediction (`TrajPred`) - Train on generated trajectories of a system, then predict unknown trajectories for new initial states.\n    - E.g: What is the behavior of this pendulum if I start from the point $[1,-1]$?\n- General and reusable - supports data from any dynamical system.\n- Novel error function Average Normalized Absolute Error (ANAE) for visualizing performance.\n- Extensive options and a ready-to-use *hyperparameter search module* to improve performance.\n- Built using [Pytorch](https://pytorch.org/), supports both CPU and GPU platforms.\n\n### Why dlkoopman?\nWe bridge the gap between a) software packages that restrict the learning of a good linearizable encoding (e.g. [`pykoopman`](https://github.com/dynamicslab/pykoopman)), and b) efforts that learn encodings for specific applications instead of being a general tool (e.g. [`DeepKoopman`](https://github.com/BethanyL/DeepKoopman)).\n\n\n## Installation\n\n### With pip (for regular users)\n`pip install dlkoopman`\n\n### From source (for development)\n```\ngit clone https://github.com/GaloisInc/dlkoopman.git\ncd dlkoopman\npip install .\n```\n\n### Running as a Docker container\nDLKoopman can also be run as a docker container by pulling the image from `galoisinc/dlkoopman:<version>`, e.g. `docker pull galoisinc/dlkoopman:v1.0.2`.\n\n\n## Tutorials and examples\nAvailable in the [`examples`](https://github.com/GaloisInc/dlkoopman/tree/ed11bef92b90112d9ca90722942a6789e6af7d5a/examples) folder.\n\n\n## Documentation and API Reference\nAvailable at https://galoisinc.github.io/dlkoopman/.\n\n\n## Description \n\n### Koopman theory\nAssume a dynamical system $x_{i+1} = F(x_i)$, where $x$ is the (genrally multi-dimensional) state of the system at index $i$, and $F$ is the (generally nonlinear) evolution rule describing the dynamics of the system. Koopman theory attempts to *encode* $x$ into a different space $y = g(x)$ where the dynamics are linear, i.e. $y_{i+1} = Ky_i$, where $K$ is the Koopman matrix. This is incredibly powerful since the state $y_i$ at any index $i$ can be predicted from the initial state $y_0$ as $y_i = K^iy_0$. This is then *decoded* back into the original space as $x = g^{-1}(y)$.\n\nFor a thorough mathematical treatment, see [this technical report](https://arxiv.org/abs/2211.07561).\n\n### dlkoopman training\n<figure>\n<img src="https://raw.githubusercontent.com/GaloisInc/dlkoopman/ed11bef92b90112d9ca90722942a6789e6af7d5a/training_architecture.png" width=750/>\n</figure>\n\nThis is a small example with three input states $\\left[x_0, x_1, x_2\\right]$. These are passed through an encoder neural network to get encoded states $\\left[y_0, y_1, y_2\\right]$. These are passed through a decoder neural network to get $\\left[\\hat{x}_0, \\hat{x}_1, \\hat{x}_2\\right]$, and also used to learn $K$. This is used to derive predicted encoded states $\\left[\\mathsf{y}_1, \\mathsf{y}_2\\right]$, which are then passed through the same decoder to get predicted approximations $\\left[\\hat{\\mathsf{x}}_1, \\hat{\\mathsf{x}}_2\\right]$ to the original input states.\n\nErrors mimimized during training:\n- Train the autoencoder - Reconstruction `recon` between $x$ and $\\hat{x}$.\n- Train the Koopman matrix - Linearity `lin` between $y$ and $\\mathsf{y}$.\n- Combine the above - Prediction `pred` between $x$ and $\\hat{\\mathsf{x}}$.\n\n### dlkoopman prediction\n<figure>\n<img src="https://raw.githubusercontent.com/GaloisInc/dlkoopman/ed11bef92b90112d9ca90722942a6789e6af7d5a/prediction_architecture.png" width=750/>\n</figure>\n\nPrediction happens after training.\n\n(a) State prediction - Compute predicted states for new indexes such as $i\'$. This uses the eigendecomposition of $K$, so $i\'$ can be any real number - positive (forward extapolation), negative (backward extrapolation), or fractional (interpolation).\n\n(b) Trajectory prediction - Generate predicted trajectories $j\'$ for new starting states such as $x^{j\'}_0$. This uses a linear neural net layer to evolve the initial state.\n\n\n## Known issues\nSome common issues and ways to overcome them are described in the [known issues](https://github.com/GaloisInc/dlkoopman/issues?q=is%3Aissue+is%3Aclosed+label%3Aknown-issue).\n\n\n## How to cite\nPlease cite the accompanying paper:\n```\n@article{Dey2022_dlkoopman,\n    author = {Sourya Dey and Eric William Davis},\n    title = {DLKoopman: A deep learning software package for Koopman theory},\n    year = {2022},\n    note = {Submitted to 5th Annual Learning for Dynamics & Control (L4DC) Conference}\n}\n```\n\n\n## References\n- B. O. Koopman - Hamiltonian systems and transformation in Hilbert space\n- J. Nathan Kutz, Steven L. Brunton, Bingni Brunton, Joshua L. Proctor - Dynamic Mode Decomposition\n- Bethany Lusch, J. Nathan Kutz & Steven L. Brunton - Deep learning for universal linear embeddings of nonlinear dynamics\n\n\n## Distribution Statement\nThis material is based upon work supported by the United States Air Force and DARPA under Contract No. FA8750-20-C-0534. Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Air Force and DARPA. Distribution Statement A, "Approved for Public Release, Distribution Unlimited."\n',
    'author': 'Sourya Dey',
    'author_email': 'sourya@galois.com',
    'maintainer': 'Galois dlkoopman team',
    'maintainer_email': 'dlkoopman@galois.com',
    'url': 'https://github.com/GaloisInc/dlkoopman',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
